% general.pl
% predicates for general use
% Sven Hartrumpf (C) 1994


% do/2
% do(N,G) is true.
% Executes the goal G N times.
% do(+integer,@callable_term)

do(1,G) :-
  !,
  G.

do(_,G) :-
  G,
  fail.

do(N,G) :-
  M is N - 1,
  do(M,G).


% doall/1
% doall(G) is true.
% Executes the goal G until if first fails.
% doall(callable_term)

doall(G) :-
  G,
  fail.

doall(_).


% list_predicate/1
% list_predicate(P) is true.
% Lists all clauses for the predicate P. P should have the follwing structure:
% name/arity.
% list_predicate(+atom/integer)

list_predicate(P/N) :-
  functor(H,P,N),
  clause(H,B),
  portray_clause((H:-B)),
  fail.

list_predicate(_) :-
  nl.


% memberv_check/2
% memberv_check(T,L) is true iff
%   the term T is member of the list L.
% T may be a variable.
% memberv_check(@term,@list)

memberv_check(Var,[H|_]) :-
  Var == H,
  !.

memberv_check(Var,[_|T]) :-
  memberv_check(Var,T).


% name_variables/1
% name_variables(V) is true.
% Names all unbound variables for displaying and the like.
% The first variable is A, the last Z. Singletons are named '_'.
% name_variables(-list)

name_variables(Vars) :-
  name_variables2(Vars,0'A).

name_variables2([],_).

name_variables2([Var|Vars],Name) :-
  var(Var),
  !,
  (
    memberv_check(Var,Vars)
  ->
    name(Var,[Name]),
    New_name is Name + 1
  ;
    Var = '_',
    New_name = Name
  ),
  name_variables2(Vars,New_name).

name_variables2([_|Vars],Name) :-
  name_variables2(Vars,Name).


% variables_in_term/2
% variables_in_term(T,V) is true iff
%   V is the list of variables in the term T.
% variables_in_term(+term,-list)

variables_in_term(T,V) :-
  variables_in_term2(T,V,[]).

variables_in_term2(T,V,V_open) :-
  var(T),
  !,
  V = [T|V_open].

variables_in_term2(T,V,V_open) :-
  T =.. [_|Arguments],
  variables_in_term_list(Arguments,V,V_open).

variables_in_term_list([],X,X).

variables_in_term_list([Term|Terms],V1,V2_open) :-
  variables_in_term2(Term,V1,V1_open),
  variables_in_term_list(Terms,V1_open,V2_open).


% write_string/1
% write_string(L) is true iff
%   the list L of character codes is written to standard output.
% write_string(+list) 

write_string([]).

write_string([H|T]) :-
  put(H),
  write_string(T).


% output predicates

% wrn/1
% wrn(T) is true iff
%   - write(T) is true, and
%   - nl is true.
% wrn(@term)

wrn(T) :-
  write(T),
  nl.


% wrl/1
% wrl(L) is true iff
%   write(L) is true for all elements in the list L.
% wrl(@list)

wrl([]).

wrl([H|T]) :-
  write(H),
  wrl(T).


% wrln/1
% wrln(L) is true iff
%   - wrl(L) is true, and
%   - nl is true.
% wrln(@list)

wrln(L) :-
  wrl(L),
  nl.


% dgoal/1
% dgoal(G) is true iff
%   - debug_on(true) is false, or
%   - the goal G succeeds.
% dgoal(+callable_term)

dgoal(Goal) :-
  (
    debug_on(true)
  ->
    Goal
  ;
    true
  ).


% dwr/1
% dwr(T) is true iff
%   - debug_on(true) is false, or
%   - write(T) is true.
% dwr(@term)

dwr(T) :-
  (
    debug_on(true)
  ->
    write(T)
  ;
    true
  ).


% dwrn/1
% dwrn(T) is true iff
%   - debug_on(true) is false, or
%   - write(T) is true, and
%   - nl is true.
% dwrn(@term)

dwrn(T) :-
  (
    debug_on(true)
  ->
    write(T),
    nl
  ;
    true
  ).


% dwrl/1
% dwrl(L) is true iff
%   - debug_on(true) is false, or
%   - wrl(L) is true.
% dwrl(@list)

dwrl(L) :-
  (
    debug_on(true)
  ->
    wrl(L)
  ;
    true
  ).


% dwrln/1
% dwrln(L) is true iff
%   - debug_on(true) is false, or
%   - wrl(L) is true, and
%   - nl is true.
% dwrln(@list)

dwrln(L) :-
  (
    debug_on(true)
  ->
    wrl(L),
    nl
  ;
    true
  ).
