% ibl.pl
% Generates the .ext, .for, .sem files from a .pre file.
% project: IBL
% Sven Hartrumpf (C) 1994


debug_on(false).

:- op(300,fx,~).
:- op(300,fx,\/).
:- op(500,xfy,&).

:- dynamic(iBl_ext_file/1).
:- dynamic(iBl_for_file/1).
:- dynamic(iBl_sem_file/1).


:- ensure_loaded('isobi_bp.pl').

% my own libraries

:- ensure_loaded('general.pl').
:- ensure_loaded('lists.pl').

% some auxiliary predicates

:- ensure_loaded('ibl_aux.pl').


% iBl_retrieve_path/3
% iBl_retrieve_path(F,P,V) is true iff
%   V is the value of the path P in the feature structure F.
%   Returns for disjunctions a value that cannot be further instantiated.
% iBl_retrieve_path(+term,+term,-term)

iBl_retrieve_path(fs(Class,F),Path,Value) :-
  iBl_path(Class,Path,Pos,_),
  iBl_retrieve(F,Pos,Value).


iBl_retrieve(Fs,[H],V) :-
  nth1(H,Fs,Value),
  iBl_partial_value(Value,V).
  
iBl_retrieve(Fs,[H|T],V) :-
  nth1(H,Fs,Value),
  iBl_retrieve(Value,T,V).


% iBl_retrieve_path0/3
% iBl_retrieve_path0(F,P,V) is true iff
%   V is the value of the path P in the feature structure F.
%   Returns for disjunctions a value that can be further instantiated.
% iBl_retrieve_path0(+term,+term,-term)

iBl_retrieve_path0(fs(Class,F),Path,Value) :-
  iBl_path(Class,Path,Pos,_),
  iBl_retrieve0(F,Pos,Value).


iBl_retrieve0(Fs,[H],V) :-
  nth1(H,Fs,Value),
  iBl_partial_value0(Value,V).
  
iBl_retrieve0(Fs,[H|T],V) :-
  nth1(H,Fs,Value),
  iBl_retrieve0(Value,T,V).


% iBl_partial_value/2
% iBl_partial_value(X,V) is true iff
%   V is the most restricted (most recent) value in X.
% iBl_partial_value(@term,-term)

iBl_partial_value(X,R) :-
  var(X),
  !,
  R = X.

iBl_partial_value(disjunction(V,X),R) :-
  var(X),
  !,
  R = V.

iBl_partial_value(disjuntion(_,single(V)),V) :-
  !.

iBl_partial_value(disjunction(_,X),V) :-
  !,
  iBl_partial_value(X,V).

iBl_partial_value(X,X).


% iBl_partial_value0/2
% iBl_partial_value0(X,V) is true iff
%   V is the most restricted (most recent) value in X.
%   V is a complex term of the form disjunction(D,Variable) if the value
%   is only partially instantiated.
% iBl_partial_value0(@term,-term)

iBl_partial_value0(X,R) :-
  var(X),
  !,
  R = X.

iBl_partial_value0(disjunction(V,X),R) :-
  var(X),
  !,
  R = disjunction(V,X).

iBl_partial_value0(disjunction(_,single(V)),V) :-
  !.

iBl_partial_value0(disjunction(_,X),V) :-
  !,
  iBl_partial_value0(X,V).

iBl_partial_value0(X,X).


% iBl_list_disjunction_value/2
% iBl_list_disjunction_value(L,V) is true iff
%   L is the list of values in the disjunction value V.
% iBl_list_disjunction_value(+list,-term)
% iBl_list_disjunction_value(-list,+term)

iBl_list_disjunction_value([H],single(H)) :-
  !.

iBl_list_disjunction_value(L,disjunction(L,_)).


% iBl_simplify_feature_structure/2
% iBl_simplify_feature_structure(F,G) is true iff
%   G is the simplified version of the feature structure F. Simplified means,
%   that only the most restricted values of disjunctions are included in the
%   feature structure.
% iBl_simplify_feature_structure(+term,-term)

iBl_simplify_feature_structure(fs(Class,F), fs(Class,G)) :-
  iBl_class_type(Class, Type),
  iBl_simplify_feature_structure2(Type, F, G).

iBl_simplify_feature_structure2(_, V, V) :-
  var(V),
  !.

iBl_simplify_feature_structure2(_, [], []) :-
  !.

iBl_simplify_feature_structure2(Type, [F1|F], [G1|G]) :-
  is_list(Type),
  !,
  Type = [H|T],
  iBl_simplify_feature_structure2(H, F1, G1),
  iBl_simplify_feature_structure2(T, F, G).

iBl_simplify_feature_structure2(_, F, G) :-
  iBl_partial_value0(F,G).


% write_fs/1
% write_fs(F) is true.
% Prints the feature structure F in a readable way.
% write_fs(@term)

write_fs(fs(Class,F)) :-
  wrl(['fs(',Class,',[ ']),
  write_fs_aux(Class,F,[1]),
  write(' ])').


write_fs_aux(_,[],_).

write_fs_aux(Class,[V|Vs],Pos) :-
  iBl_path(Class,_,Pos,Type),
  iBl_complex_type(Type, _),
  !,
  append(Pos,[1],New_pos),
  write('[ '),
  write_fs_aux(Class,V,New_pos),
  write('] '),
  append(Pos1,[N],Pos),
  M is N + 1,
  append(Pos1,[M],Pos2),
  write_fs_aux(Class,Vs,Pos2).

write_fs_aux(Class,[V|Vs],Pos) :-
  iBl_path(Class,Path,Pos,Type),
  path_end(Path,Feature_name),
  write(Feature_name),
  write(': '),
  write_atomic_value(Type,V),
  (
    Vs \== [] ->
      write('  ')
    ;
      true
  ),
  append(Pos1,[N],Pos),
  M is N + 1,
  append(Pos1,[M],Pos2),
  write_fs_aux(Class,Vs,Pos2).


write_atomic_value(_,V) :-
  var(V),
  !,
  write(V).

write_atomic_value(string_type,V) :-
  !,
  (
    is_list(V)
  ->
    write_string(V)
  ;
    write('complex '),write(V)
  ).

write_atomic_value(_,\/List) :-
  !,
  write('{'),
  write_disjunction(List),
  write('}').

write_atomic_value(_,V) :-
  write(V).


write_disjunction([H]) :-
  !,
  write(H).

write_disjunction([H|T]) :-
  write(H),
  write(' '),
  write_disjunction(T).


% write_compact_fs/1
% write_compact_fs(F) is true.
% Prints the feature structure F in a compact representation that Prolog can
% read as a term.
% write_compact_fs(@term)

write_compact_fs(fs(Class,F)) :-
  write('('),
  write(Class),
  write(',['),
  write_compact_fs_aux(Class,F,[1]),
  write('])').


write_compact_fs_aux(_,[],_).

write_compact_fs_aux(Class,[V|Vs],Pos) :-
  iBl_path(Class,_,Pos,Type),
  iBl_complex_type(Type, _),
  !,
  append(Pos,[1],New_pos),
  write('['),
  write_compact_fs_aux(Class,V,New_pos),
  write(']'),
  (
    Vs \== [] ->
      write(',')
    ;
      true
  ),
  append(Pos1,[N],Pos),
  M is N + 1,
  append(Pos1,[M],Pos2),
  write_compact_fs_aux(Class,Vs,Pos2).

write_compact_fs_aux(Class,[V|Vs],Pos) :-
  iBl_path(Class,_,Pos,Type),
  write_compact_atomic_value(Type,V),
  (
    Vs \== [] ->
      write(',')
    ;
      true
  ),
  append(Pos1,[N],Pos),
  M is N + 1,
  append(Pos1,[M],Pos2),
  write_compact_fs_aux(Class,Vs,Pos2).


write_compact_atomic_value(_,V) :-
  var(V),
  !,
  write(V).

write_compact_atomic_value(string_type,V) :-
  !,
  (
    is_list(V)
  ->
    write('"'),
    write_string(V),
    write('"')
  ;
    write('complex '),write(V)
  ).

write_compact_atomic_value(_,\/List) :-
  !,
  write('disjunction('),
  write(List),
  write(',_)').

write_compact_atomic_value(_,V) :-
  write(V).


% path_end/2
% path_end(+P,?E)
%   E is the smallest nonempty suffix of the path P.
% path_end(+term,-atom)

path_end(_:T,End) :-
  !,
  path_end(T,End).

path_end(E,E).


% iBl_all_general_extensions/0
% iBl_all_general_extensions is true.
% Generates the general extension of all lexical classes and stores the result.

iBl_all_general_extensions :-
  doall((
    iBl_pclass(Class,lexical,_,_,_,_),
    % !, % only for testing one class
    iBl_general_extension(Class,E),
    iBl_store_general_extension(Class,E)
  )).


% iBl_general_extension/2
% iBl_general_extension(C,E) is true iff
%   E is the list of the global parts of all feature structures in the general
%   extension of class C.
% iBl_general_extension(+term,-list)

iBl_general_extension(Class,E) :-
  iBl_general_extension(Class,E,[]).


iBl_general_extension(Class,E,Conditions) :-
  dwrln(['general extension of class ',Class]),
  iBl_pclass(Class,_,Cpl,_,_,_),
  dwrln(['CPL: ',Cpl]),
  last_element(Cpl,Category),
  iBl_new_object(Category,Object),
  iBl_apply_conditions(Conditions,Object),
  dwrln(['general_extension initial object: ',Object]),
  iBl_general_extension_aux([Class|Cpl],[Object],E0),
  iBl_remove_local_part(E0, E),
  dwrln(['The general extension of class ',Class,' is:']),
  dgoal((process_list(X^(write_fs(X),nl),E),nl)).


iBl_general_extension_aux([],Result,Result).

iBl_general_extension_aux([Class|Classes],E,Result) :-
  dwrln(['superclass extension regarding ',Class]),
  iBl_superclass_extensions(E,Class,E1),
  iBl_general_extension_aux(Classes,E1,Result).


% iBl_superclass_extensions/3
% iBl_superclass_extensions(O,C,R) is true iff
%   R is the list of all feature structures in the superclass extension of
%   all feature structures in O with respect to the class C.
% iBl_superclass_extensions(+list,+term,-list)

iBl_superclass_extensions([],_,[]).

iBl_superclass_extensions([H|T],Class,List) :-
  iBl_superclass_extension(H,Class,List1),
  iBl_superclass_extensions(T,Class,List2),
  append(List1,List2,List).


% iBl_superclass_extension/3
% iBl_superclass_extension(O,C,R) is true iff
%   R is the list of all feature structures in the superclass extension of
%   the feature structure O with respect to the class C.
% iBl_superclass_extension(+term,+term,-list)

iBl_superclass_extension(Object,Superclass,Result) :-
  dwrln(['superclass_extension object: ',Object]),
  iBl_main_extension(Object,Superclass),
  dwrln(['superclass_extension main extension object: ',Object]),
  iBl_default_extension(Object,Superclass),
  dwrln(['superclass_extension default extension object: ',Object]),
  iBl_variant_extension(Object,Superclass,Result),
  dwrln(['superclass_extension variant extension object: ',Result]).


% iBl_main_extension/2
% iBl_main_extension(O,C) is true iff
%   The feature structure O satisies the condition stated in the main equation
%   set of the class C.
% The feature structure O might be further instantiated by this predicate.
% iBl_main_extension(-term,@term)

iBl_main_extension(O,Class) :-
  iBl_pclass(Class,_,_,Main,_,_),
  dwrln(['main_extension main equation set: ',Main]),
  iBl_apply_conditions(Main,O).


% iBl_default_extension/2
% iBl_default_extension(O,C) is true iff
%   The feature structure O satisies the condition stated in the default equation
%   set of the class C.
% The feature structure O might be further instantiated by this predicate.
% iBl_default_extension(-term,@term)

iBl_default_extension(O,Class) :-
  iBl_pclass(Class,_,_,_,Default,_),
  dwrln(['default_extension default equation set: ',Default]),
  iBl_apply_default_conditions(Default,O).


% iBl_variant_extension/3
% iBl_default_extension(O,C,R) is true iff
%   The list R contains all feature structures that are calculated by applying
%   one variant equation set to the feature structure O.
% iBl_default_extension(@term,@term,-list)

iBl_variant_extension(O,Class,New_objects) :-
  iBl_pclass(Class,_,_,_,_,Variants),
  dwrln(['variant_extension variant equation sets: ',Variants]),
  iBl_apply_condition_list(Variants,O,New_objects).


iBl_apply_condition_list([],O,[O]) :-
  !.

iBl_apply_condition_list(L,O,Result) :-
  iBl_apply_condition_list2(L,O,Result).


iBl_apply_condition_list2([],_,[]).

iBl_apply_condition_list2([H|T],O,List) :-
  copy_term(O,New_object),
  dwrln(['apply_condition_list2 conditions: ',H]),
  iBl_apply_conditions(H,New_object),
  !,
  dwrln(['apply_condition_list2 result: ',New_object]),
  iBl_apply_condition_list2(T,O,List1),
  List = [New_object | List1].

iBl_apply_condition_list2([_|T],O,New_objects) :-
  iBl_apply_condition_list2(T,O,New_objects).


% iBl_new_object/2
% iBl_new_object(C,O) is true iff
%   O is a feature structure of the type of class C whose values are all
%   variables. 
% iBl_new_object(@term,-term)

iBl_new_object(Class,fs(Class,Fs)) :-
  iBl_class_type(Class,Type),
  var_structure(Type,Fs).


% var_structure/2
% var_structure(T,S) is true iff
%   S has the same structure of nested lists as T, but contains only variables.
% var_structure(+term,-term)

var_structure([],[]).

var_structure([H|T],[_|Rest]) :-
  atomic(H),
  !,
  var_structure(T,Rest).

var_structure([H|T],[H1|T1]) :-
  var_structure(H,H1),
  var_structure(T,T1).


% iBl_apply_default_conditions/2
% iBl_apply_default_conditions(C,O) is true.
% The feature structure O might be further instantiated in order to satisfy the
% conditions in the list C.
% iBl_apply_default_conditions(+list,-term)

iBl_apply_default_conditions([],_).

iBl_apply_default_conditions([Condition|Conditions],O) :-
  iBl_apply_default_condition(Condition,O),
  iBl_apply_default_conditions(Conditions,O).


iBl_apply_default_condition(Condition,O) :-
  %dwrln(['apply_default_condition condition: ',Condition]),
  iBl_apply_condition(Condition,O),
  !.

iBl_apply_default_condition(_,_).


% iBl_apply_conditions/2
% iBl_apply_conditions(C,O) is true iff
%   the feature structure O satisfies all conditions in the list C.
% The feature structure O might be further instantiated in order to satisfy the
% conditions in the list C.
% iBl_apply_conditions(+list,-term)

iBl_apply_conditions([],_).

iBl_apply_conditions([Condition|Conditions],O) :-
  iBl_apply_condition(Condition,O),
  iBl_apply_conditions(Conditions,O).



% iBl_apply_condition/2
% iBl_apply_condition(C,O) is true iff
%   the feature structure O satisfies the condition C.
% The feature structure O might be further instantiated in order to satisy C.
% iBl_apply_condition(@term,-term)

iBl_apply_condition(L = R,Object) :-
  !,
  (
    iBl_retrieve_path0(Object,L,L_value) ->
      true
  ;
    term_to_codes(L,E0),
    append(E0,"This path doesn't exist.",E),
    iBl_error(E)
  ),
  iBl_apply_condition2(L,L_value,R,Object).

iBl_apply_condition(Predicate0, Object) :-
  dwrln(['apply_condition Predicate0: ',Predicate0]),
  Predicate0 =.. [Functor|Args0],
  iBl_predicate_arguments(Object,Args0,Args),
  Predicate =.. [Functor|Args],
  dwrln(['apply_condition predicate to call: ',Predicate]),
  call(Predicate).


% string condition

iBl_apply_condition2(L,L_value,R,Object) :-
  Object = fs(Class,_),
  iBl_path(Class,L,_,string_type),
  !,
  dwrln(['string condition L_value: ',L_value]),
  iBl_string_condition(L_value,R,Object),
  dwrln(['string condition new L_value: ',L_value]).

% not a string condition

iBl_apply_condition2(L,L_value,R,Object) :-
  %dwrln(['apply_condition2 L_value: ',L_value]),
  iBl_evaluate_right_side(Object,L,R,R_value),
  %dwrln(['apply_condition2 R_value: ',R_value]),
  iBl_equation(L_value,R_value).


%

iBl_string_condition(L_value,E & Rest,Object) :-
  !,
  iBl_concat_to_list(Object,E & Rest,Elements),
  dwrln(['string_condition right side elements: ',Elements]),
  iBl_string_concat(Elements,L_value).

iBl_string_condition(L_value,R,_) :-
  is_list(R),
  !,
  L_value = R.

iBl_string_condition(L_value,R,Object) :-
  iBl_retrieve_path(Object,R,R_value),
  L_value = R_value.


%

iBl_string_concat(Elements,Result) :-
  var(Result),
  !,
  iBl_string_construct(Elements,Result),
  dwrln(['string_concat construct result: ',Result]).

iBl_string_concat(Elements,Result) :-
  % \+ var(Result),
  iBl_string_decompose(Result,Elements).


%

iBl_string_construct([A|Rest],Result) :-
  iBl_string_construct2(A,Rest,Result).

iBl_string_construct2(A,[],A).

iBl_string_construct2(A,[H|T],R) :-
  is_list(A),
  is_list(H),
  !,
  append(A,H,R1),
  iBl_string_construct2(R1,T,R).

iBl_string_construct2(A,[H|T],R) :-
  iBl_string_construct2(concat(A,H),T,R).


iBl_string_decompose(String,[A,B]) :-
  !,
  append(A,B,String).

iBl_string_decompose(String,[H|T]) :-
  append(H,New_string,String),
  iBl_string_decompose(New_string,T).

%

iBl_concat_to_list(Object,E & Rest,[Element|Elements]) :-
  !,
  %dwrln(['concat_to_list E: ']),
  iBl_concat_to_list2(Object,E,Element),
  %dwrln(['concat_to_list Element: ',Element]),
  iBl_concat_to_list(Object,Rest,Elements).

iBl_concat_to_list(Object,E,[Element]) :-
  iBl_concat_to_list2(Object,E,Element).

iBl_concat_to_list2(_,T,R) :-
  var(T),
  !,
  R = T.

iBl_concat_to_list2(_,[Char|Chars],[Char|Chars]).

iBl_concat_to_list2(O,Feature,Value) :-
  % add type check (must be string)
  iBl_retrieve_path(O,Feature,Value).


% iBl_evaluate_right_side/4
% iBl_evaluate_right_side(O,L,V,R) is true iff
%   the value V of the right side of an equation, whose left side has type L,
%   evaluates to R for the feature structure O.
% iBl_evaluate_right_side(+term,+term,+term,-term)

% string

iBl_evaluate_right_side(_,_,[Char|Chars],[Char|Chars]) :-
  !.

% path

iBl_evaluate_right_side(O,_,Feature,Value) :-
  iBl_retrieve_path0(O,Feature,Value),
  !.

% negation of atomic feature structures

iBl_evaluate_right_side(fs(Class,_),L,~Value,R) :-
  !,
  iBl_path(Class,L,_,Type),
  dwrln(['evaluate_right_side negation of ',Value,' of type ',Type]),
  iBl_atomic_type(Type,Values),
  value_of_type(Value,Type),
  list_subtraction(Values,[Value],R1),
  (
    R1 = [H]
  ->
    R = H
  ;
    R = disjunction(R1,_)
  ).

% disjunction of atomic feature structures

iBl_evaluate_right_side(fs(Class,_),L,Values \/ Value,R) :-
  !,
  iBl_path(Class,L,_,Type),
  iBl_atomic_type(Type,Allowed_values),
  iBl_disjunction_to_list(Allowed_values,Values \/ Value,R1),
  R = disjunction(R1,_).

% atomic feature structure

iBl_evaluate_right_side(fs(Class,_),L,Value,R) :-
  iBl_path(Class,L,_,Type),
  iBl_atomic_type(Type,_),
  (
    value_of_type(Value,Type)
  ->
    R = Value
  ;
    term_to_codes(Value,E0),
    append("The value ",E0,E1),
    append(E1," is not of the type ",E2),
    term_to_codes(Type,E3),
    append(E2,E3,E4),
    append(E4,".",E),
    iBl_error(E)
  ).


% iBl_disjunction_to_list/3
% iBl_disjunction_to_list(A,D,L) is true iff
%   - the list L contains exactly the values in the disjunction D, and
%   - all values in D are in the list A of allowed values.
% iBl_disjunction_to_list(@list,@term,-list)

iBl_disjunction_to_list(Allowed_values,D,L) :-
  iBl_disjunction_to_list_aux(Allowed_values,D,[],L).

iBl_disjunction_to_list_aux(Allowed_values,Rest \/ Last,L,Result) :-
  !,
  iBl_disjunction_to_list_aux(Allowed_values,Rest,[Last|L],Result).

iBl_disjunction_to_list_aux(Allowed_values,T,L,Result) :-
  member_check(T,Allowed_values),
  !,
  Result = [T|L].


% iBl_equation/2
% iBl_equation(L,R) is true iff
%   L can be unified with the value R.
% Note: for disjunctions, this can result in further instantiating.
% iBl_equation(-term,-term)

iBl_equation(L,R) :-
  var(L),
  !,
  L = R.

% The disjunction values for constants on the RHS are instantiated further.
% That's not necessary, but it doesn't matter, and one needs fewer clauses. 

iBl_equation(disjunction(L,X),disjunction(R,X)) :-
  !,
  list_intersection(L,R,N),
  N \== [],
  iBl_list_disjunction_value(N,X).

iBl_equation(disjunction(L,X),R) :-
  !,
  %dwrln(['equation ',L,' and ',R]),
  member_check(R,L),
  iBl_list_disjunction_value([R],X).

iBl_equation(L,disjunction(R,X)) :-
  !,
  %dwrln(['equation ',L,' and ',R]),
  member_check(L,R),
  iBl_list_disjunction_value([L],X).

iBl_equation(L,R) :-
  L == R.


% iBl_predicate_arguments/3
% iBl_predicate_arguments(O,A,N) is true iff
%   N is the list A of arguments with all values that are paths replaced by the
%   corresponding values in the feature structure O.
% iBl_predicate_arguments(@term,+list,-list)

iBl_predicate_arguments(_,[],[]).

iBl_predicate_arguments(Object,[Arg|Args],[New_arg|New_args]) :-
  (
    iBl_retrieve_path0(Object,Arg,New_arg)
  ->
    true
  ;
    New_arg = Arg
  ),
  iBl_predicate_arguments(Object,Args,New_args).


% iBl_remove_local_part/2
% iBl_remove_local_part(L, N) is true iff
%   N is the list L of feature structures without the local parts of the feature
%   structures.
% iBl_remove_local_part(+list,-list)

iBl_remove_local_part([], []).

iBl_remove_local_part([fs(Class,F)|Rest], [fs(Class,Global)|Globals]) :-
  iBl_class_global(Class,F,Global0),
  dwrln(['remove_local_part: Global0: ',Global0]),
  iBl_simplify_feature_structure(fs(Class,Global0), fs(Class,Global)),
  dwrln(['remove_local_part: Global: ',Global]),
  iBl_remove_local_part(Rest, Globals).


% iBl_store_general_extension/2
% iBl_store_general_extension(C,E) is true iff
%   The general extension E of the lexical class C is stored in the .ext,
%   .for, (.sem) files. E is a list of feature structures generated by
%   iBl_general_extension/2.
% iBl_store_general_extension(+term,+list)

iBl_store_general_extension(_,[]) :-
  !.

iBl_store_general_extension(Class,E) :-
  metacall(iBl_ext_file(Ext_file)),
  tell(Ext_file),
  iBl_store_general_extension_ext(E),
  metacall(iBl_for_file(For_file)),
  tell(For_file),
  iBl_store_general_extension_for(Class,E),
  tell(user).


% iBl_store_general_extension_ext/1
% iBl_store_general_extension_ext(E) is true iff
%   The general extension E is stored in the .ext-file. E is a list of feature
%   structures generated by iBl_general_extension/2.
% Variables in the feature structure are instantiated to atoms in oder to have
%  more readable names in the .ext-file.
% iBl_store_general_extension_ext(-term)

iBl_store_general_extension_ext([]).

iBl_store_general_extension_ext([fs(Class,F)|Es]):-
  variables_in_term(F,V),
  name_variables(V),
  write('iBl_entry'),
  write_compact_fs(fs(Class,F)),
  wrn('.'),
  iBl_store_general_extension_ext(Es).
  

% iBl_store_general_extension_for/2
% iBl_store_general_extension_for(C,E) is true iff
%   The word forms in the general extension E of the lexical class C are written
%   to the .for-file. The format is as follows: length of the word form (1 byte),
%   the numerical identifier C (3 bytes), the word form itself.
% iBl_store_general_extension_for(+term,+list)

iBl_store_general_extension_for(Class,E) :-
  setof(Form,Form^Fs^(member(Fs,E),iBl_retrieve_path(Fs,form,Form)),List),
  X3 is Class mod 256,
  Class1 is Class // 256,
  X2 is Class1 mod 256,
  X1 is Class1 // 256,
  doall((
    member(Form,List),
    length(Form,Form_length),
    put(Form_length),
    put(X1),
    put(X2),
    put(X3),
    write_string(Form)
    )).


% iBl_process/1
% iBl_process(N) is true iff
%   - the file N.pre can be successfully processed as a preprocessed IBL lexicon
%     file, and
%   - the result is written to a .ext- and a .for-file.
% iBl_process(+atom)

iBl_process(N) :-
  statistics(runtime,[T1,_]),
  atom_concat(N,'.ext',Ext_file),
  assertz(iBl_ext_file(Ext_file)),
  atom_concat(N,'.for',For_file),
  assertz(iBl_for_file(For_file)),
  wrn('generating general extensions'),
  iBl_all_general_extensions,
  wrn('writing the global parts of all general extensions'),
  tell(Ext_file),
  told,
  tell(For_file),
  told,
  statistics(runtime,[T2,_]),
  T is (T2 - T1) / 1000,
  wrln(['run time: ',T,' s']), 
  wrn('ibl finished').


% predicate for testing

t :-
  iBl_process(german).
