% ibl_aux.pl
% auxiliary predicates for the IBL system
% project: IBL
% Sven Hartrumpf (C) 1994


% list_of_variables/2
% list_of_variables(N,L) is true iff
%   L is a list of N new variables.
% list_of_variables(+integer,-list)

list_of_variables(0,[]) :-
  !.

list_of_variables(N,[_|L]) :-
  M is N - 1,
  list_of_variables(M,L).


% term_to_codes/2
% term_to_codes(T,S) is true iff
%   S is a list of codes representing the term S.
% term_to_codes(+term,?list)

term_to_codes(T,C) :-
  var(T),
  !,
  C = T.

term_to_codes(T,C) :-
  T =.. [Functor|Args],
  atom_codes(Functor,Functor_codes),
  %write('term_to_codes functor '),write(Functor_codes),nl,
  term_list_to_codes(Args,Args_codes),
  %write('term_to_codes arguments '),write(Args_codes),nl,
  term_to_codes_aux(Functor_codes,Args_codes,C).

% must be extended (prints all operators in prefix notation)

term_to_codes_aux(Functor,[],Functor) :-
  !.

term_to_codes_aux(Functor,Args,S) :-
  append(Functor,"(",S0),
  term_to_codes_aux2(Args,S1),
  append(S0,S1,S2),
  append(S2,")",S).

term_to_codes_aux2([S],S) :-
  !.

term_to_codes_aux2([S0|T],S) :-
  !,
  append(S0,",",S1),
  term_to_codes_aux2(T,S2),
  append(S1,S2,S).


term_list_to_codes([],[]).

term_list_to_codes([Term|Terms],[C|Rest]) :-
  term_to_codes(Term,C),
  term_list_to_codes(Terms,Rest).


% list_goal/2
% list_goal(L,G) is true iff
%   G is a goal that contains all goals in the list L.
% list_goal(+list,-callable_term)
% list_goal(-list,+callable_term)

list_goal([Goal],Goal) :-
  \+ functor(Goal,',',2),
  !.

list_goal([H|T],(Goal,Goals)) :-
  list_goal([H],Goal),
  list_goal(T,Goals).


% parse_ops/3
% parse_ops(P,T,R) is true iff
%   R is a list of elements of the form (Keyword,Value), where Keyword is
%   a xfy operator with precedence P that is found in the term T. Value is the
%   second operand of the operator Keyword. 
% parse_ops(+integer,@term,-list)

parse_ops(Precedence,I,[(Keyword,Value)|Result]) :-
  arg(2,I,Arg),
  functor(I,Keyword,2),
  current_op(Precedence,xfy,Keyword),
  arg(1,Arg,Value),
  functor(Arg,Next_keyword,2),
  current_op(Precedence,xfy,Next_keyword),
  !,
  parse_ops(Precedence,Arg,Result).

parse_ops(Precedence,I,[(Keyword,Arg)]) :-
  arg(2,I,Arg),
  functor(I,Keyword,2),
  current_op(Precedence,xfy,Keyword),
  !.

  
% part_value/3
% part_value(K,P,V) is true iff
%   V is the value for the keyword K in the parse result P.
% part_value(+atom,@list,?atom)

part_value(K,P,none) :-
  \+ part_value_aux(K,P,_),
  !.

part_value(K,P,V) :-
  part_value_aux(K,P,V).

part_value_aux(K,[(K,V)|_],V).

part_value_aux(K,[_|T],V) :-
  part_value_aux(K,T,V).


% value_of_type/2
% value_of_type(V,T) is true iff
%   T is the name of a type that has V as one of its values.
% Succeeds for every type that fulfills this condition once.
% value_of_type(+atom,?atom)
% value_of_type(?atom,+atom)

value_of_type(Value,Type) :-
  metacall(iBl_atomic_type(Type,Values)),
  member_check(Value,Values).


% iBl_error/1
% iBl_error(S) is true.
% Prints the string S as an IBL error.
% iBl_error(+list)

iBl_error(S) :-
  nl,
  write('IBL error:'),
  nl,
  write('  '),
  write_string(S),
  nl,
  nl,
  fail.
