% iblp.pl
% program for preprocessing IBL lexicons
% project: IBL (a lexicon formalism using multiple default inheritance)
% Sven Hartrumpf (C) 1994


debug_on(false).


% dynamic predicates (only for BinProlog necessary)

:- dynamic(iBl_atomic_type/2).
:- dynamic(iBl_class/6).
:- dynamic(iBl_class_global/3).
:- dynamic(iBl_class_id/2).
:- dynamic(iBl_classes/1).
:- dynamic(iBl_class_type/2).
:- dynamic(iBl_complex_type/2).
:- dynamic(iBl_path/4).
:- dynamic(iBl_pclass/6).
:- dynamic(iBl_predicate/1).


% define operators

:- op(1070,fx,class).
:- op(1070,fx,top).
:- op(1070,fx,word).
:- op(1060,xfy,global).
:- op(1060,xfy,inherit).
:- op(1060,xfy,local).
:- op(1060,xfy,main).
:- op(1060,xfy,default).
:- op(1060,xfy,variant).
:- op(750,fx,type).
:- op(750,fx,include).
:- op(300,fx,~).
:- op(300,fx,\/).
:- op(500,xfy,&).


% come closer to the ISO standard regarding built-in predicates

:- ensure_loaded('isobi_bp.pl').

% my own libraries

:- ensure_loaded('general.pl').
:- ensure_loaded('lists.pl').

% some auxiliary predicates

:- ensure_loaded('ibl_aux.pl').


% iBl_assert/1
% iBl_assert(T) is true iff
%   T is a valid lexicon term.
% iBl_assert(@term)

% declaration of a simple type

iBl_assert(type Name = {Values0}) :-
  !,
  list_goal(Values,Values0),
  dwrln(['simple type ',Name,': ',Values]),
  (
    iBl_type_name(Name)
  ->
    atom_codes(Name,E0),
    append("type ",E0,E1),
    append(E1," defined twice.",E),
    iBl_error(E)
  ;
    assertz(
      iBl_atomic_type(Name,Values)
    )
  )
.

% declaration of a complex type

iBl_assert(type Name = Elements0) :-
  !,
  list_goal(Elements,Elements0),
  dwrln(['complex type ',Name,': ',Elements]),
  (
    iBl_type_name(Name)
  ->
    atom_codes(Name,E0),
    append("type ",E0,E1),
    append(E1," defined twice.",E),
    iBl_error(E)
  ;
    assertz(
      iBl_complex_type(Name,Elements)
    )
  )
.

% top class

iBl_assert((top Class global X)) :-
  !,
  iBl_parse_top(Class,(dummy global X)).

% clauses for a nonlexical class

iBl_assert((class Class inherit Superclasses0 main X)) :-
  !,
  list_goal(Superclasses,Superclasses0),
  iBl_parse_class(Class,Superclasses,(dummy main X)).

iBl_assert((class Class inherit Superclasses0 default X)) :-
  !,
  list_goal(Superclasses,Superclasses0),
  iBl_parse_class(Class,Superclasses,(dummy default X)).

iBl_assert((class Class inherit Superclasses0 variant X)) :-
  !,
  list_goal(Superclasses,Superclasses0),
  iBl_parse_class(Class,Superclasses,(dummy variant X)).

iBl_assert((class Class main X)) :-
  !,
  iBl_parse_class(Class,[],(dummy main X)).

iBl_assert((class Class default X)) :-
  !,
  iBl_parse_class(Class,[],(dummy default X)).

iBl_assert((class Class variant X)) :-
  !,
  iBl_parse_class(Class,[],(dummy variant X)).

% clauses for a lexical class

iBl_assert((word Class inherit Superclasses0 main X)) :-
  !,
  list_goal(Superclasses,Superclasses0),
  iBl_parse_word(Class,Superclasses,(dummy main X)).

iBl_assert((word Class inherit Superclasses0 default X)) :-
  !,
  list_goal(Superclasses,Superclasses0),
  iBl_parse_word(Class,Superclasses,(dummy default X)).

iBl_assert((word Class inherit Superclasses0 variant X)) :-
  !,
  list_goal(Superclasses,Superclasses0),
  iBl_parse_word(Class,Superclasses,(dummy variant X)).

% predicate definition

iBl_assert(X) :-
  functor(X,Functor,Arity),
  dwrln(['asserting lexicon term with functor ',Functor/Arity,
         ' as a Prolog term.']),
  assertz(iBl_predicate(X)).


% iBl_parse_top/2
% iBl_parse_top(C,X) is true iff
%   The class C with body X is a valid top class.
% Stores the result in the Prolog database and produces a description of the
% feature structure type.
% iBl_parse_top(+term,@term)
 
iBl_parse_top(Class,X) :-
  dwrln(['parse top class ',Class,'.']),
  parse_ops(1060,X,Parts),
  dwrln(['parts: ',Parts]),
  % check keywords in Parts
  part_value(global,Parts,G0),
  list_goal(G,G0),
  part_value(local,Parts,L0),
  (
    L0 \== none
  ->
    list_goal(L,L0)
  ;
    L = []
  ),
  part_value(main,Parts,M0),
  (
    M0 \== none
  ->
    list_goal(M,M0)
  ;
    M = []
  ),
  part_value(default,Parts,D0),
  (
    D0 \== none
  ->
    list_goal(D,D0)
  ;
    D = []
  ),
  dwrln(['main    : ',M]),
  dwrln(['default : ',D]),
  findall(List,(part_value(variant,Parts,V),V \== none,list_goal(List,V)),Vs),
  dwrln(['variants: ',Vs]),
  iBl_store_class(Class,nonlexical,[],M,D,Vs,Class_id),
  iBl_make_class_type(Class_id,G,L).


% iBl_parse_word/3
% iBl_parse_word(C,S,X) is true iff
%   The class C with body X and the superclasses S is a valid lexical class.
% Stores the result in the Prolog database. 
% iBl_parse_word(+term,+list,@term)

iBl_parse_word(Class,Superclasses,X) :-
  write_string(Class),
  nl,
  parse_ops(1060,X,Parts),
  part_value(main,Parts,M0),
  (
    M0 \== none
  ->
    list_goal(M,M0)
  ;
    M = []
  ),
  part_value(default,Parts,D0),
  (
    D0 \== none
  ->
    list_goal(D,D0)
  ;
    D = []
  ),
  dwrln(['main    : ',M]),
  dwrln(['default : ',D]),
  findall(List,(part_value(variant,Parts,V),V \== none,list_goal(List,V)),Vs),
  dwrln(['variants: ',Vs]),
  iBl_store_class(Class,lexical,Superclasses,M,D,Vs,_).


% iBl_parse_class/3
% iBl_parse_class(C,S,X) is true iff
%   The class C with body X and the superclasses S is a valid nonlexical class.
% Stores the result in the Prolog database. 
% iBl_parse_class(+term,+list,@term)

iBl_parse_class(Class,Superclasses,X) :-
  dwrln(['parse class ',Class,'.']),
  parse_ops(1060,X,Parts),
  dwrln(['parts : ',Parts]),
  % check keywords in Parts
  part_value(main,Parts,M0),
  (
    M0 \== none
  ->
    list_goal(M,M0)
  ;
    M = []
  ),
  part_value(default,Parts,D0),
  (
    D0 \== none
  ->
    list_goal(D,D0)
  ;
    D = []
  ),
  dwrln(['main    : ',M]),
  dwrln(['default : ',D]),
  findall(List,(part_value(variant,Parts,V),V \== none,list_goal(List,V)),Vs),
  dwrln(['variants: ',Vs]),
  iBl_store_class(Class,nonlexical,Superclasses,M,D,Vs,_).


% iBl_store_class/7
% iBl_store_class(C,T,S,M,D,V,N) is true iff
%   the class C with type T (lexical or nonlexical), superclasses S, main
%   equation set M, default equation set D, variant equation sets V can be
%   stored as the N-th class in the Prolog database using the dynamic predicates
%   iBl_class/6 and iBl_class_id/2.
% The number of classes stored in the fact iBl_classes/1 is increased by one.
% iBl_store_class(+term,+list,+list,+list,+list,+list,?integer)

iBl_store_class(Class,Type,Superclasses,M,D,Vs,N) :-
  metacall(iBl_classes(N0)),
  N is N0 + 1,
  assertz(iBl_class(N,Type,Superclasses,M,D,Vs)),
  assertz(iBl_class_id(Class,N)),
  retract(iBl_classes(N0)),
  assertz(iBl_classes(N)).


% iBl_make_class_type/3
% iBl_make_class_type(C,G,L) is true iff
%   G is a valid global part and L is a valid local part of a feature structure
%   type for the top class C.
% The result is stored in the Prolog database using the dynamic predicates
% iBl_class_type/2, iBl_class_global/3, and iBl_path/4.
% iBl_make_class_type(+term,+term,+term)

iBl_make_class_type(Class,Global,Local) :-
  dwrln(['make_class_type global : ',Global]),
  check_feature_structure_type(Global),
  iBl_make_class_type_aux(Class,Global,[],[1],S_global),
  %dwrln(['make_class_type S_global : ',S_global]),
  length(S_global,N0),
  N is N0 + 1,
  list_of_variables(N0,Global_part),
  append(Global_part,_,Match),
  assertz(iBl_class_global(Class,Match,Global_part)),
  (
    local == none
  -> 
    S_local = []
  ;
    check_feature_structure_type(Local),
    iBl_make_class_type_aux(Class,Local,[],[N],S_local)
  ),
  append(S_global,S_local,S),
  assertz(iBl_class_type(Class,S)).


iBl_make_class_type_aux(_,[],_,_,[]) :-
  !.

iBl_make_class_type_aux(Class,Feature:Feature_type,Path_elements,Pos,
  Feature_type) :-
  metacall(iBl_atomic_type(Feature_type,_)),
  !,
  reverse_list([Feature|Path_elements],Path_elements0),
  iBl_list_path(Path_elements0,Path),
  assertz(iBl_path(Class,Path,Pos,Feature_type)).

iBl_make_class_type_aux(Class,Feature:Feature_type,Path_elements,Pos,S) :-
  dwrln(['make_class_type_aux ',Feature,' : ',Feature_type]),
  metacall(iBl_complex_type(Feature_type,Type)),
  !,
  Path_elements0 = [Feature|Path_elements],
  reverse_list(Path_elements0,Path_elements1),
  iBl_list_path(Path_elements1,Path),
  assertz(iBl_path(Class,Path,Pos,Feature_type)),
  append(Pos,[1],New_pos),
  iBl_make_class_type_aux(Class,Type,Path_elements0,New_pos,S).

iBl_make_class_type_aux(_,_:Feature_type,_,_,_) :-
  !,
  term_to_codes(Feature_type,E1),
  append(E1," is an unknown type.",E),
  iBl_error(E).

iBl_make_class_type_aux(Class,[First|Rest],Path,Pos,S) :-
  !,
  iBl_make_class_type_aux(Class,First,Path,Pos,S_first),
  append(Pos1,[N],Pos),
  M is N + 1,
  append(Pos1,[M],New_pos),
  iBl_make_class_type_aux(Class,Rest,Path,New_pos,S_rest),
  S = [S_first|S_rest].

iBl_make_class_type_aux(_,X,_,_,_) :-
  term_to_codes(X,E1),
  append("incorrect type definition: ",E1,E),
  iBl_error(E).


% iBl_list_path/2
% iBl_list_path(L,P) is true iff
%   P is a path connecting all elements in L with the infix operator ^.
% iBl_list_path(+list,?term)
% iBl_list_path(-list,+term)

iBl_list_path([H|T],H^P) :-
  T \== [],
  !,
  iBl_list_path(T,P).

iBl_list_path([H],H).


% iBl_type_name/1
% iBl_type_name(T) is true iff
%   T is the name of a type.
% iBl_type_name(?atom)

iBl_type_name(Name) :-
  (
    metacall(iBl_atomic_type(Name,_))
  ->
    true
  ;
    metacall(iBl_complex_type(Name,_))
  ).


% check_feature_name/1
% check_feature_name(N) is true iff
%   N is not the name of a value of an existing type.
% check_feature_name(+term)

check_feature_name(Name) :-
  value_of_type(Name,Type),
  !,
  atom_codes(Name,E0),
  append("The feature name ",E0,E1),
  append(E1," is not allowed as it is a value of the type ",E2),
  atom_codes(Type,E3),
  append(E2,E3,E4),
  append(E4,".",E),
  iBl_error(E).

check_feature_name(_).


% check_feature_structure_type/1
% check_feature_structure_type(T) is true iff
%   - T is a list of pairs feature:type, and
%   - no feature name in T is the name of a value, and
%   - all feature types in T are defined.
% check_feature_structure_type(@list)

check_feature_structure_type([]) :-
  !.

check_feature_structure_type([Feature:Type|Rest]) :-
  !,
  check_feature_name(Feature),
  (
    iBl_type_name(Type)
  ->
    true
  ;
    term_to_codes(Type,E0),
    append("Undefined type: ",E0,E),
    iBl_error(E)
  ),
  check_feature_structure_type(Rest).

check_feature_structure_type(T) :-
  term_to_codes(T,E0),
  append("Not a valid feature structure type: ",E0,E),
  iBl_error(E).


% iBl_produce_cpls/0
% iBl_produce_cpls is true iff
%   every lexical class has a valid CPL.
% The result is stored in the Prolog database using the dynamic predicate
% iBl_pclass/6, while the corresponding iBl_class/6 clause is retracted.
% iBl_produce_cpls/0

iBl_produce_cpls :-
  doall((
    metacall(iBl_class(Class,Type,_,M,D,V)),
    cpl_of_class(Class,Cpl),
    (
      Type == lexical
    ->
      iBl_valid_lexical_cpl(Cpl,_)
    ;
      true
    ),
    retract(iBl_class(Class,_,_,_,_,_)),
    assertz(iBl_pclass(Class,Type,Cpl,M,D,V)))
  ).


% cpl_of_class/2
% cpl_of_class(C,Cpl) is true iff
%   Cpl is the CPL of the class C.
% cpl_of_class(+term,?list)

cpl_of_class(Class,Cpl) :-
  metacall(iBl_pclass(Class,_,Cpl,_,_,_)),
  !.

cpl_of_class(Class,Cpl) :-
  metacall(iBl_class(Class,_,Superclasses,_,_,_)),
  iBl_class_ids(Superclasses,S),
  %dwrln(['cpl_of_class superclass ids : ',S]),
  cpl_of_classes([],S,List),
  remove_front_doubles(List,Cpl).


cpl_of_classes(L,[],L).

cpl_of_classes(List,[H|T],List2) :-
  cpl_of_class(H,H_cpl),
  append(List,[H|H_cpl],List1),
  cpl_of_classes(List1,T,List2).


% iBl_class_ids/2
% iBl_class_ids(C,R) is true iff
%   R is the list of numerical identifiers for all class names in the list.
% iBl_class_ids(+list,?list)

iBl_class_ids([],[]).

iBl_class_ids([Class|Classes],[Id|Ids]) :-
  (
    metacall(iBl_class_id(Class,Id))
  ->
    true
  ;
    iBl_classname_to_codes(Class,E1),
    append("The class ",E1,E2),
    append(E2," is not defined.",E),
    iBl_error(E)
  ),
  iBl_class_ids(Classes,Ids).


% iBl_valid_lexical_cpl/2
% iBl_valid_lexical_cpl(L,C) is true iff
%   - L is a valid CPL for a lexical class, and
%   - C is the top class in the CPL.
% iBl_valid_lexical_cpl(+list,?term)

iBl_valid_lexical_cpl([Class],Class) :-
  !,
  (
    metacall(iBl_class_type(Class,_))
  ->
    true
  ;
    iBl_classname_to_codes(Class,E0),
    append("There is no top class in the CPL of the lexical class ",E0,E1),
    append(E1,".",E),
    iBl_error(E)
  ).

iBl_valid_lexical_cpl([Class|Classes],Top) :-
  (
    metacall(iBl_class_type(Class,_))
  ->
    iBl_classname_to_codes(Class,E0),
    append("There is more than one top class in the CPL of the lexical class ",
           E0,E1),
    append(E1,".",E),
    iBl_error(E)
  ;
    true
  ),
  iBl_valid_lexical_cpl(Classes,Top).


% iBl_classname_to_codes/2
% iBl_classname_to_codes(C,N)
%   N is a list of character codes representing the name of class C.
% iBl_classname_to_codes(+atom,?character_code_list)
% iBl_classname_to_codes(+character_code_list,?character_code_list)

iBl_classname_to_codes(C,R) :-
  atom(C),
  !,
  atom_codes(C,R).

iBl_classname_to_codes(C,C).


% pure dynamic predicates

% iBl_atomic_type/2
% iBl_atomic_type(T,V) is true iff
%   V is the list of possible values for the atomic type T.
% iBl_atomic_type(?atom,?list)


% iBl_class/6
% iBl_class(N,T,S,M,D,Vs) is true iff
%   The class with the numerical identifier Id is of type T, has the
%   superclasses in the list S, the main equation set M, the default equation
%   set D, and the variant equation sets Vs.
% iBl_class(?integer,?atom,?list,?list,?list,?list) 


% iBl_class_global/3
% iBl_class_global(C,F1,F2) is true iff
%   F2 are the features that are in F1 and global in the top class C.
% iBl_class_global(+integer,+list,?list)


% iBl_class_id/2
% iBl_class_id(C,I) is true iff
%   I is the numerical identifier of the class whose name is C.
% iBl_class_id(?term,?integer)


% iBl_classes/1
% iBl_classes(N) is true iff
%   N is the number of classes stored in the Prolog database.
% iBl_classes(?integer)


% iBl_class_type/2
% iBl_class_type(C,T) is true iff
%   The nested list structure T describes the feature structure type of the
%   top class C.
% iBl_class_type(?integer,?list)


% iBl_complex_type/2
% iBl_complex_type(T,V) is true iff
%   V is the list of possible values for the atomic type T.
% iBl_complex_type(?atom,?list)


% iBl_path/4
% iBl_path(C,P,Pos,T) is true iff
%   in class C the path P is stored at the position Pos and has the type T.
% iBl_path(?integer,?term,?list,?atom)


% iBl_pclass/6
% iBl_pclass(Id,T,Cpl,M,D,Vs) is true iff
%   The class with the numerical identifier Id is of type T, has the CPL Cpl,
%   the main equation set M, the default equation set D, and the variant equation
%   sets Vs.
% iBl_pclass(?integer,?atom,?list,?list,?list,?list) 


% iBl_predicate/1
% iBl_predicate(C) is true iff
%   C is a clause defining a predicate that will be copied to the preprocessed
%   lexicon.
% iBl_predicate(?clause)


% end of pure dynamic predicates


% iBl_predicates/1
% iBl_predicates(L) is true iff
%   L is the list of all dynamic predicates used in this program.
% iBl_predicates(?list)

iBl_predicates([iBl_atomic_type/2,iBl_class/6,iBl_class_global/3,
  iBl_class_id/2,iBl_classes/1,iBl_class_type/2,iBl_complex_type/2,
  iBl_path/4,iBl_pclass/6,iBl_predicate/1]).


% iBl_list/0
% iBl_list is true iff
%   All predicates that are defined in iBl_predicates/1 are printed to standard
%   output.
% iBl_list

iBl_list :-
  iBl_predicates(Predicates),
  process_list(X^(list_predicate(X)),Predicates).


% iBl_new/0
% iBl_new is true iff
%   - all predicates that are defined in iBl_predicates/1 are abolished, and
%   - the standard types are stored, and
%   - the number of classes is set to 0.
% iBl_new

iBl_new :-
  iBl_predicates(Predicates),
  process_list(X^(abolish(X)),Predicates),
  assertz(iBl_atomic_type(boolean_type,[+,-])),
  assertz(iBl_atomic_type(general_type,complex)),
  assertz(iBl_atomic_type(string_type,complex)),
  assertz(iBl_classes(0)).


% iBl_load/1
% iBl_load(N) is true iff
%   The file with the name N can be successfully loaded as a .ibl-lexicon.
% iBl_load(+atom)

iBl_load(Filename) :-
  atom_concat(Filename,'.ibl',N),
  iBl_load2(N, [], Files),
  dwrn(Files).


% iBl_load2/3
% iBl_load2(N, R1, R2) is true iff
%   - The file with the name N is contained in the list of already loaded files
%     R1 or can be successfully loaded as a .ibl-lexicon, and
%   - R2 is the list of loaded files after loading the file N.
% iBl_load2(+atom,+list,-list)

iBl_load2(N, Read_files, Read_files) :-
  member_check(N, Read_files),
  !.

iBl_load2(N, Read_files, New_read_files) :-
  see(N),
  wrln(['> reading ',N]),
  iBl_load3([N|Read_files], New_read_files),
  wrln(['> finished reading ',N]),
  seen,
  !.

iBl_load2(N,Files,Files) :-
  atom_codes(N,E0),
  append("Unable to complete reading file ",E0,E),
  iBl_error(E).


iBl_load3(Read_files, New_read_files) :-
  read(Clause),
  (
    compound(Clause),
    Clause = (include Name)
  ->
    dwrln(['load3: include file:',Name]),
    seeing(Current),
    iBl_load2(Name, Read_files, Read_files1),
    see(Current),
    iBl_load3(Read_files1, New_read_files)
  ;
    Clause == end_of_file
  ->
    New_read_files = Read_files
  ;
    iBl_assert(Clause),
    iBl_load3(Read_files, New_read_files)
  ).


% iBl_write_embedded_predicates/0
% iBl_write_embedded_predicates is true.
% All clauses that are stored using the dynamic predicate ibl_predicate/1 are
% written to standard output with meaningful names.
% iBl_write_embedded_predicates

iBl_write_embedded_predicates :-
  doall((
    metacall(iBl_predicate(T)),
    variables_in_term(T, Vars),
    name_variables(Vars),
    wrln([T,'.'])
  )).


% iBl_save_pre_file/1
% iBl_save_pre_file(N) is true iff
%   the preprocessed lexicon file N.pre is successfully written.
% iBl_save_pre_file(+atom)

iBl_save_pre_file(Filename) :-
  atom_concat(Filename,'.pre',N),
  tell(N),
  write('% This is a preprocessed lexicon file automatically generated from an IBL-file.'),
  nl,nl,
  write(':- ensure_loaded(''ibl.pl'').'),
  nl,nl,
  iBl_write_embedded_predicates,
  nl,
  process_list(X^(list_predicate(X)),[iBl_atomic_type/2,iBl_complex_type/2,
    iBl_class_global/3,iBl_class_type/2,iBl_path/4,iBl_pclass/6]),
  told.


% iBl_preprocess/1
% iBl_preprocess(N) is true iff
%   the IBL lexicon file N.ibl is preprocessed and successfully written to
%   the file N.pre.
% iBl_preprocess(+atom)

iBl_preprocess(N) :-
  statistics(runtime,[T1,_]),
  iBl_new,
  iBl_load(N),
  dwrn('first pass complete'),
  dwrn('producing CPLs'),
  iBl_produce_cpls,
  wrn('writing the preprocessed lexicon'),
  iBl_save_pre_file(N),
  statistics(runtime,[T2,_]),
  T is (T2 - T1) / 1000,
  wrln(['run time: ',T,' s']), 
  wrn('iblp finished.').


% predicate for testing

t :-
  iBl_preprocess(german).
